<?php defined('SYSPATH') OR die('No direct access allowed.');

require_once('op5/log.php');

class Synergy_Model extends Model
{
	protected static $_synergy_path = "/opt/synergy/";

	private static function log($level, $msg) {
		$log = Op5Log::instance('synergy');
		$log->log($level, $msg);
	}

	/**
	 *	Fetch a list of all saved trees
	 *	@return array
	 */
	static function getRuleList() {
		$result = array();
		$rules_dir = self::$_synergy_path . "etc/bps";
		if (!is_dir($rules_dir)) {
			throw new SynergyException("$rules_dir is not a directory.");
		}
		if (false === ($dh = opendir($rules_dir))) {
			throw new SynergyException("Can not open directory $rules_dir.");
		}
		while(false !== ($file = readdir($dh))) {
			if(($file[0] === ".") || pathinfo($file, PATHINFO_EXTENSION) !== "lua") continue;
			$file = urldecode($file);
			$result[] = substr($file, 0, -4);
		}
		natcasesort($result);
		// Javascript prints the DOM with reverse sort for some reason. Will find out why later
		return array_reverse($result, true);
	}

	/**
	 * Checks if logged in user is able to see BSM object based on the
	 * service.
	 *
	 * The method is public for testability.
	 */
	static function __is_permitted_by_services($data, $auth) {
		if ($auth->authorized_for('service_view_all')) {
			return true;
		}

		// Group rules shouldn't authorize for limited users if not materialized
		if ($data["result"]["service_contacts"] === "*"
				&& (!isset($data['items']) || !isset($data['materialized']))) {
			return true;
		}

		// MON-11399: Prevent PHP error in future if 'service_contacts' should ever be '*' here instead of array.
		if (!is_array($data["result"]["service_contacts"])) {
			self::log('error', "'service_contacts' for {$data['name']} is expected to be an array but is not");
			return false;
		}

		// Users is contact for service
		if (in_array($auth->get_user()->get_username(), $data["result"]["service_contacts"], true)) {
			return true;
		}

		return false;
	}

	/**
	 * Checks if logged in user is able to see BSM object based on the
	 * host.
	 *
	 * The method is public for testability.
	 */
	static function __is_permitted_by_hosts($data, $auth) {
		if ($auth->authorized_for('host_view_all')) {
			return true;
		}

		// Group rules shouldn't authorize for limited users if not materialized
		if ($data["result"]["host_contacts"] === "*"
				&& (!isset($data['items']) || !isset($data['materialized']))) {
			return true;
		}

		// MON-11399: Prevent PHP error in future if 'service_contacts' should ever be '*' here instead of array.
		if (!is_array($data["result"]["host_contacts"])) {
			self::log('error', "'host_contacts' for {$data['name']} is expected to be an array but is not");
			return false;
		}

		// User is contact for host
		if (in_array($auth->get_user()->get_username(), $data["result"]["host_contacts"], true)) {
			return true;
		}
		return false;
	}

	/**
	 *	Pretty printing JSON
	 *	@param $block array
	 *	@param $indent int
	 *	@return string
	 */
	public static function _block_format(array $block, $ident = 1) {
		$result = "{\n";
		$i = str_repeat("\t", $ident);
		$c = 0;
		$size = sizeof($block);
		foreach($block as $key => $value) {
			$sep = ++$c == $size ? "" : ",";
			if(is_array($value)) {
				$add = "";
				if(!is_numeric($key)) {
					$add = $key . " = \n" . str_repeat("\t", $ident);
				}
				$result .= sprintf("$i$add%s$sep\n", self::_block_format($value, $ident + 1));
				continue;
			}
			if($key === "intent") {
				$result .= sprintf("$i%s = %s$sep\n", $key, $value);
				continue;
			}
			$result .= sprintf("$i%s = \"%s\"$sep\n", $key, $value);
		}
		return $result . str_repeat("\t", $ident - 1). "}";
	}

	/**
	 *	Saves created rule.
	 *	@param $rule array
	 *	@param $rule_name string
	 *	@param $sid string
	 *	@throws Exception on failure
	 */
	public static function saveRule(array $rule, $rule_name, $sid) {
		$file_name = urlencode($rule_name);
		$dir_path = self::$_synergy_path . "var/queue/$sid";
		if (!is_dir($dir_path)) {
			// Sets an resets the umask to be able to create a directory with sufficient permissions
			$old_umask = umask(0);
			if(!mkdir($dir_path, 0770, true)) {
				throw new Exception(sprintf("Queue directory '%s' does not exist, and I cannot create it", $dir_path));
			}
			umask($old_umask);
		}
		$filename = "$dir_path/$file_name.lua";
		if(!@file_put_contents($filename, "script = " . self::_block_format($rule))) {
			throw new Exception(sprintf("Could not save BSM rule file to queue: '%s'", $filename));
		}
	}

	/**
	 *	Move saved rules to spool directory
	 *	@param $sid string
	 *	@throws Exception
	 */
	public static function commitActions($sid)
	{
		$auth = Auth::instance();
		$dir_path = self::$_synergy_path . "var/queue/$sid";
		if (!is_dir($dir_path)) {
			$msg = "Directory " . $dir_path . " doesn't exist";
			self::log('error', $msg);
			throw new Exception($msg);
		}
		if (!$auth->authorized_for('export')) {
			$msg = $auth->get_user()->get_username() . ' is not allowed to save configuration.';
			self::log('error', $msg);
			throw new Exception($msg);
		}
		exec("mv -f $dir_path ".self::$_synergy_path."var/spool/", $output, $return_code);
		if($return_code) {
			throw new Exception(sprintf("Could not move '%s' to '%s'", $dir_path, self::$_synergy_path."var/spool/"));
		}
		// Force a new loop to verify rules and to process current states
		// Throws exception on failure
		self::daemonRequest('forceloop');
	}

	/**
	 *	Deletes saved rule
	 *	@param $rule_name string
	 *	@return bool
	 */
	public static function deleteRule($rule_name) {
		$file_name = urlencode($rule_name);
		$rule_path = self::$_synergy_path . "etc/bps/$file_name.lua";
		if (!file_exists($rule_path)) {
			self::log('error', 'Rule ' . $rule . ' wanted for deletion does not exist. Huh, an error with desired result...');
			return false;
		}
		if(!@unlink($rule_path)) {
			self::log('error', 'Rule ' . $rule . ' wanted for deletion, but file could not be deleted.');
			return false;
		}
		return true;
	}

	/**
	 *	Deletes list of saved rules
	 *	@param $list array
	 *	@param $sid string
	 *	@throws Exception
	 */
	public static function deleteRulesList($list, $sid) {
		foreach ($list as $key => $rule) {
			$file_name = urlencode($rule);
			if (!file_exists(self::$_synergy_path . "etc/bps/".$file_name.".lua")) {
				self::log('error', 'Rule ' . $rule . ' wanted for deletion does not exist. Huh, an error with desired result...');
				continue;
			}
			$list[$key] = $file_name;
		}
		$dir_path = self::$_synergy_path . "var/queue/$sid";
		if (!is_dir($dir_path)) {
			$old_umask = umask(0);
			if(!mkdir($dir_path, 0770, true)) {
				umask($old_umask);
				throw new Exception(sprintf("Queue directory '%s' does not exist, and I cannot create it", $dir_path));
			}
			umask($old_umask);
		}
		$filename = "$dir_path/delete.list";
		if(!file_put_contents($filename, json_encode(array_values($list)))) {
			throw new Exception(sprintf("Could not write to queue dirs for deletion at '%s'", $filename));
		}
	}

	/**
	 * Send a command to daemon
	 *
	 * @param $cmd string
	 * @return string JSON if $cmd is something like "top" or "full", a plain
	 * string if $cmd is "forceloop" or some other command. See bin/synergy
	 * @throws SynergySocketException
	 **/
	static function daemonRequest($cmd) {
		/* FIXME: don't hardcode the synergy socket path */
		$socket = @fsockopen('unix:///opt/monitor/var/rw/synergy', NULL, $errno, $errstr);
		if(!$socket) {
			self::log('error', 'Error connecting to daemon socket ('.$errno.'): ' . $errstr);
			throw new SynergySocketException("Could not connect to Synergy daemon. Is the synergy daemon running? Try to run: \"service synergy restart\" from the op5 monitor system command line");
		}

		fwrite($socket, $cmd . "\n");
		$json = "";
		do {
			$response = "";
			$response = fread($socket, 2048);
			$json .= $response;
		} while ($response !== "");
		fclose($socket);
		return $json;
	}

	/**
	 * Json decode that catches errors with the decoding
	 *
	 * @param $json string
	 * @throws SynergyException
	 * @return array
	 **/
	function decode_json_to_array($json)
	{
		$data = json_decode($json, true);
		if ($data === null) {
			$jerr = json_last_error();
			self::log('error', 'Invalid return from json_decode() (' . var_export($data, true) . "). error=" . $jerr . ", input=" . $json);
			throw new SynergyException("json_decode() failed");
		}
		return $data;
	}

	/**
	 * Fetch structure and states for all trees
	 *
	 * @throws Exception
	 * @return array
	 */
	static function getStates() {
		$json = self::daemonRequest("full");
		$data = self::decode_json_to_array($json);
		$result = self::auth_tree_recursive($data);
		uksort($result['items'], 'strnatcasecmp');
		$result['items'] = array_reverse($result['items'], true);
		return $result;
	}

	/**
	 * Recursively auth object in Synergy results
	 *
	 * @param $tree array
	 * @return array
	 **/
	static function auth_tree_recursive($tree)
	{
		$auth = Auth::instance();
		foreach ($tree['items'] as $key => $item) {
			unset($tree['items'][$key]);
			// Either host contacts or service contacts gives auth for group rules
			if (isset($item['items']) && (self::__is_permitted_by_hosts($item, $auth) || self::__is_permitted_by_services($item, $auth))) {
				// New group rule, only follow this branch if parent is authed
				$item = self::auth_tree_recursive($item);
				$tree['items'][urldecode($key)] = $item;
			} else if (self::__is_permitted_by_hosts($item, $auth) && self::__is_permitted_by_services($item, $auth)) {
				$tree['items'][urldecode($key)] = $item;
			}
		}
		return $tree;
	}

	/**
	 * Fetch top level results.
	 *
	 * @todo: Uses some sort of funky authentication which must be overseen
	 * @throws SynergyException
	 * @return array
	 */
	static function getTopsList() {
		$json = self::daemonRequest("top");
		$data = self::decode_json_to_array($json);
		$auth = Auth::instance();
		$result = array();
		foreach($data['items'] as $rule => $content) {
			if(self::__is_permitted_by_services($content, $auth)) {
				$result[urldecode($rule)] = $content;
			}
		}
		uksort($result, 'strnatcasecmp');
		return array_reverse($result, true);
	}

	/**
	 *	Checks if given rule exists
	 *	@param $rule string
	 *	@return bool
	 */
	public static function ruleExists($rule) {
		$rule = urlencode($rule);
		return file_exists(self::$_synergy_path."etc/bps/$rule.lua");
	}

	/**
	 *	Fetches the different models that can be used to create rules. i.e best_state etc.
	 *	@return string
	 */
	public static function getActionsJson() {
		$actions_path = self::$_synergy_path . "etc/rules/repo";
		if(!is_dir($actions_path))
			throw new Exception("The rule template directory doesn't exist." . $actions_path);
		if(!($dir = opendir($actions_path)))
			throw new Exception("Cannot open rule template directory. " . $actions_path);
		$result = array();
		while(($file = readdir($dir)) !== false) {
			if($file[0] !== "." && pathinfo($file, PATHINFO_EXTENSION) === "lua") {
				$result[] = substr($file, 0, -4);
			}
		}
		$custom_actions_path = self::$_synergy_path . "etc/rules/custom/repo";
		if(!is_dir($custom_actions_path)) return false;
		if(!($dir = opendir($custom_actions_path))) return false;
		while(($file = readdir($dir)) !== false) {
			if($file[0] !== "." && array_pop(explode(".", $file)) === "lua") {
				$result[] = substr($file, 0, -4);
			}
		}
		$actions = addslashes(json_encode(array_unique($result)));
		$json = array();
		$status = 0;
		exec("echo $actions | " . self::$_synergy_path . "bin/actions --all 2>&1", $json, $status);
		if($status !== 0) {
			self::log("error", "[$status] Syntax error in a rule file: ".var_export($json, true));
			return false;
		}
		return implode("",$json);
	}

	/**
	 *	Fetch all nagios object names of type (hosts, services, hostgroups, serviegroups) or ALL objects if no type is specified
	 *	@param $object_type string
	 *	@param $encoded bool
	 *	@return array
	 */
	public static function getNagiosObjectsJson($object_type = false, $encoded = true) {
		$ls = Livestatus::instance();
		$result = array();
		switch ($object_type) {
			case 'hosts':
				$hosts = $ls->getHosts(array('columns'=>array('name')));
				foreach ($hosts as $host) {
					$result[$host['name']] = $host['name'];
				}
				break;
			case 'services':
				$services = $ls->getServices(array('columns'=>array("host_name", "description")));
				foreach ($services as $host_desc) {
					$host = $host_desc['host_name'];
					$desc = $host_desc['description'];
					if (!isset($result[$host])) {
						$result[$host] = array();
					}
					array_push($result[$host], $desc);
				}
				break;
			case 'hostgroups':
				$hostgroups = $ls->getHostgroups(array('columns'=>array("name")));
				foreach ($hostgroups as $hostgroup) {
					$result[$hostgroup['name']] = $hostgroup['name'];
				}
				break;
			case 'servicegroups':
				$servicegroups = $ls->getServicegroups(array('columns'=>array("name")));
				foreach ($servicegroups as $servicegroup) {
					$result[$servicegroup['name']] = $servicegroup['name'];
				}
				break;
			default:
				// Fetch_all. This shouldn't be used for synergy. Fetch on demand instead
				$result['hosts'] = self::getNagiosObjectsJson('hosts', false);
				$result['services'] = self::getNagiosObjectsJson('services', false);
				$result['hostgroups'] = self::getNagiosObjectsJson('hostgroups', false);
				$result['servicegroups'] = self::getNagiosObjectsJson('servicegroups', false);
				return json_encode($result);
				break;
		}
		if (!$encoded) {
			return $result;
		} else {
			return json_encode($result);
		}
	}
}
